/* AXPbox Alpha Emulator
 * Copyright (C) 2020 Tomáš Glozar
 * Website: https://github.com/lenticularis39/axpbox
 *
 * Forked from: ES40 emulator
 * Copyright (C) 2007-2008 by the ES40 Emulator Project
 * Copyright (C) 2007 by Camiel Vanderhoeven
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 *
 * Although this is not required, the author would appreciate being notified of,
 * and receiving any modifications you may make to the source code that might
 * serve the general public.
 *
 * Parts of this file based upon the Poco C++ Libraries, which is Copyright (C)
 * 2004-2006, Applied Informatics Software Engineering GmbH. and Contributors.
 */

//
// Bugcheck.h
//
// $Id: Bugcheck.h,v 1.1 2008/05/31 15:47:20 iamcamiel Exp $
//
// Library: Foundation
// Package: Core
// Module:  Bugcheck
//
// Definition of the Bugcheck class and the self-testing macros.
//
// Copyright (c) 2004-2006, Applied Informatics Software Engineering GmbH.
// and Contributors.
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
//
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//

#ifndef Foundation_Bugcheck_INCLUDED
#define Foundation_Bugcheck_INCLUDED

#include "Foundation.hpp"
#include <string>

class CBugcheck
/// This class provides some static methods that are
/// used by the
/// poco_assert_dbg(), poco_assert(), poco_check_ptr()
/// and poco_bugcheck() macros.
/// You should not invoke these methods
/// directly. Use the macros instead, as they
/// automatically provide useful context information.
{
public:
  static void assertion(const char *cond, const char *file, int line);
  /// An assertion failed. Break into the debugger, if
  /// possible, then throw an AssertionViolationException.

  static void nullPointer(const char *ptr, const char *file, int line);
  /// An null pointer was encountered. Break into the debugger, if
  /// possible, then throw an NullPointerException.

  static void bugcheck(const char *file, int line);
  /// An internal error was encountered. Break into the debugger, if
  /// possible, then throw an BugcheckException.

  static void bugcheck(const char *msg, const char *file, int line);
  /// An internal error was encountered. Break into the debugger, if
  /// possible, then throw an BugcheckException.

  static void debugger(const char *file, int line);
  /// An internal error was encountered. Break into the debugger, if
  /// possible.

  static void debugger(const char *msg, const char *file, int line);
  /// An internal error was encountered. Break into the debugger, if
  /// possible.

protected:
  static std::string what(const char *msg, const char *file, int line);
};

//
// useful macros (these automatically supply line number and file name)
//
#define poco_assert(cond)                                                      \
  if (!(cond))                                                                 \
    CBugcheck::assertion(#cond, __FILE__, __LINE__);                           \
  else                                                                         \
    (void)0

#define poco_check_ptr(ptr)                                                    \
  if (!(ptr))                                                                  \
    CBugcheck::nullPointer(#ptr, __FILE__, __LINE__);                          \
  else                                                                         \
    (void)0

#define poco_bugcheck() CBugcheck::bugcheck(__FILE__, __LINE__)

#define poco_bugcheck_msg(msg) CBugcheck::bugcheck(msg, __FILE__, __LINE__)

#define poco_debugger() CBugcheck::debugger(__FILE__, __LINE__)

#define poco_debugger_msg(msg) CBugcheck::debugger(msg, __FILE__, __LINE__)

#endif // Foundation_Bugcheck_INCLUDED
